@@ -1,4 +1,4 @@
-{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE FlexibleInstances, DeriveDataTypeable #-}
 -- Full-featured argument parsing library for Haskell programs
 -- Bart Massey <bart@cs.pdx.edu>
 
@@ -30,6 +30,9 @@ module System.Console.ParseArgs (
   Arg(..),
   Argtype(..), 
   ArgsComplete(..),
+  ArgsDash(..),
+  ArgsParseControl(..),
+  APCData(..),
   -- ** DataArg and its pseudo-constructors
   DataArg,
   argDataRequired, argDataOptional, argDataDefaulted,
@@ -186,12 +189,7 @@ data (Ord a) => Args a =
 -- when argument parsing fails.  The first argument is the usage
 -- message, the second the actual error message from the parser.
 data ParseArgsException = ParseArgsException String String
-     deriving Eq
-
-instance Typeable ParseArgsException where
-    typeOf _ = mkTyConApp e [s, s] where
-        e = mkTyCon "ParseArgsException"
-        s = typeOf ""
+     deriving (Eq, Typeable)
 
 instance Exception ParseArgsException
 
@@ -223,6 +221,11 @@ arg_optional :: (Ord a) =>
 arg_optional (Arg { argData = Just (DataArg { dataArgOptional = b }) }) = b
 arg_optional _ = True
 
+arg_required :: (Ord a) =>
+                Arg a   -- ^Argument.
+             -> Bool    -- ^True if argument is required to be present.
+arg_required a = not (arg_optional a)
+
 -- |Return the value of a defaulted argument.
 arg_default_value :: (Ord a)
                   => Arg a         -- ^Argument.
@@ -241,6 +244,7 @@ arg_default_value arg@(Arg { argData = Just
 arg_default_value _ = Nothing
 
 -- |There's probably a better way to do this.
+perhaps :: Bool -> String -> String
 perhaps b s = if b then s else ""
 
 -- |Format the described argument as a string.
@@ -299,10 +303,10 @@ make_keymap :: (Ord a, Ord k, Show k) =>
                ((Arg a) -> Maybe k)   -- ^Mapping from argdesc to flag key.
             -> [ Arg a ]              -- ^List of argdesc.
             -> (Map.Map k (Arg a))    -- ^Map from key to argdesc.
-make_keymap f_field args =
+make_keymap f_field ads =
     (keymap_from_list .
      filter_keys .
-     map (\arg -> (f_field arg, arg))) args
+     map (\arg -> (f_field arg, arg))) ads
 
 -- |How \"sloppy\" the parse is.
 data ArgsComplete = ArgsComplete         -- ^Any extraneous arguments
@@ -317,16 +321,45 @@ data ArgsComplete = ArgsComplete         -- ^Any extraneous arguments
                                          -- permitted, and will be skipped,
                                          -- saved, and returned.
 
+-- |Whether to always treat an unknown argument beginning
+-- with \"-\" as an error, or to allow it to be used as a
+-- positional argument when possible.
+data ArgsDash = ArgsHardDash   -- ^If an argument begins with
+                               -- a \"-\", it will always be
+                               -- treated as an error unless
+                               -- it corresponds to a flag description.
+              | ArgsSoftDash   -- ^If an argument beginning with
+                               -- a \"-\" is unrecognized as a flag,
+                               -- treat it as a positional argument
+                               -- if possible. Otherwise it is an error.
+              deriving Eq
+
+-- |Record containing the collective parse control information.
+data ArgsParseControl = ArgsParseControl {
+  apcComplete :: ArgsComplete,
+  apcDash :: ArgsDash }
+
+-- |Class for building parse control information,
+-- for backward compatibility.
+class APCData a where
+  getAPCData :: a -> ArgsParseControl
+
+instance APCData ArgsParseControl where
+  getAPCData a = a
+
+instance APCData ArgsComplete where
+  getAPCData a = ArgsParseControl a ArgsHardDash
+
 -- |The iteration function is given a state and a list, and
 -- expected to produce a new state and list.  The function
--- is again invoked with the resulting state and list.
--- When the function returns the empty list, `exhaust` returns
--- the final state produced.
+-- is again invoked with the resulting state and list.  When
+-- the supplied function returns the empty list, this
+-- function returns the final state produced.
 exhaust :: (s -> [e] -> ([e], s))   -- ^Function to iterate.
         -> s                        -- ^Initial state.
         -> [e]                      -- ^Initial list.
         -> s                        -- ^Final state.
-exhaust f s [] = s
+exhaust _ s [] = s
 exhaust f s l =
   let (l', s') = f s l
   in exhaust f s' l'
@@ -338,30 +371,29 @@ parseError :: String    -- ^Usage message.
 parseError usage msg =
   throw (ParseArgsException usage msg)
 
--- |Given a description of the arguments, `parseArgs` produces
--- a map from the arguments to their \"values\" and some other
--- useful byproducts.  `parseArgs` requires that the argument
--- descriptions occur in the order 1) flag arguments, 2) required
--- positional arguments, 3) optional positional arguments; otherwise
--- a runtime error will be thrown.
-parseArgs :: (Show a, Ord a) =>
-             ArgsComplete   -- ^Degree of completeness of parse.
+-- |Given a description of the arguments, `parseArgs`
+-- produces a map from the arguments to their \"values\" and
+-- some other useful byproducts.  `parseArgs` requires that
+-- the argument descriptions occur in the order 1) flag
+-- arguments, then 2) positional arguments; otherwise a
+-- runtime error will be thrown.
+parseArgs :: (Show a, Ord a, APCData b) =>
+             b              -- ^Configuration for parse.
           -> [ Arg a ]      -- ^Argument descriptions.
           -> String         -- ^Full program pathname.
           -> [ String ]     -- ^Incoming program argument list.
           -> Args a         -- ^Outgoing argument parse results.
-parseArgs acomplete argd pathname argv =
+parseArgs apcData argd pathname argv =
   runST (do
            check_argd
-           let flag_args = takeWhile arg_flag argd
-           let posn_args = dropWhile arg_flag argd
+           let (flag_args, posn_args) = span arg_flag argd
            let name_hash = make_keymap argName flag_args
            let abbr_hash = make_keymap argAbbr flag_args
            let prog_name = baseName pathname
            let usage = make_usage_string prog_name
-           let (am, posn, rest) = exhaust (parse usage name_hash abbr_hash)
-                                  (Map.empty, posn_args, [])
-                                  argv
+           let (am, _, rest) = exhaust (parse usage name_hash abbr_hash)
+                                (Map.empty, posn_args, [])
+                                argv
            let required_args = filter (not . arg_optional) argd
            unless (and (map (check_present usage am) required_args))
                   (error "internal error")
@@ -385,19 +417,15 @@ parseArgs acomplete argd pathname argv =
     --- Check for various possible misuses.
     check_argd :: ST s ()
     check_argd = do
-      --- Order must be flags, posn args, optional posn args
-      let residue = dropWhile arg_flag argd
-      let residue' = dropWhile arg_fixed_posn residue
-      let residue'' = dropWhile arg_opt_posn residue'
-      unless (null residue'')
-             (argdesc_error "argument description in wrong order")
+      --- Order must be flags, then posn args
+      let (_, posns) = span arg_flag argd
+      unless (all arg_posn posns)
+             (argdesc_error "argument description mixes flags and positionals")
       --- No argument may be "nullary".
       when (or (map arg_nullary argd))
            (argdesc_error "bogus 'nothing' argument")
       return ()
       where
-        arg_fixed_posn a = (arg_posn a) && (not (arg_optional a))
-        arg_opt_posn a = (arg_posn a) && (arg_optional a)
         arg_nullary (Arg { argName = Nothing,
                            argAbbr = Nothing,
                            argData = Nothing }) = True
@@ -418,109 +446,120 @@ parseArgs acomplete argd pathname argv =
             perhaps
               (not (null posn_args))
               (" " ++ unwords (map arg_string posn_args)) ++
-            (case acomplete of
+            (case apcComplete $ getAPCData apcData of
                ArgsComplete -> ""
                ArgsTrailing s -> " [--] [" ++ s ++ " ...]"
                ArgsInterspersed -> " ... [--] ...") ++ "\n"
         --- argument lines
         arg_lines = concatMap (arg_line n) argd where
-            arg_line n a =
+            arg_line na a =
                 let s = arg_string a in
                 "  " ++ s ++ 
-                replicate (n - (length s)) ' ' ++
+                replicate (na - (length s)) ' ' ++
                 "  " ++ argDesc a ++ "\n"
     --- simple recursive-descent parser
     parse _ _ _ av@(_, _, []) [] = ([], av)
     parse usage _ _ av [] =
-        case acomplete of
+        case apcComplete $ getAPCData apcData of
           ArgsComplete -> parseError usage "unexpected extra arguments"
           _ -> ([], av)
     parse usage name_hash abbr_hash (am, posn, rest) av@(aa : aas) =
         case aa of
-          "--" -> case acomplete of
-                    ArgsComplete -> parseError usage
-                                      ("unexpected -- " ++
-                                      "(extra arguments not allowed)")
+          "--" -> case getAPCData apcData of
+                    ArgsParseControl ArgsComplete ArgsHardDash -> 
+                      parseError usage ("unexpected -- " ++
+                        "(extra arguments not allowed)")
                     _ -> ([], (am, posn, (rest ++ aas)))
-          s@('-' : '-' : name) ->
+          s@('-' : '-' : name) 
+            | isJust (Map.lookup name name_hash) ||
+              apcDash (getAPCData apcData) == ArgsHardDash ->
               case Map.lookup name name_hash of
-                Just ad -> peel s ad aas
+                Just ad -> 
+                  let (args', am') = peel s ad aas in
+                  (args', (am', posn, rest))
                 Nothing ->
-                    case acomplete of
-                      ArgsInterspersed ->
-                          (aas, (am, posn, rest ++ ["--" ++ name]))
-                      _ -> parseError usage
-                           ("unknown argument --" ++ name)
-          ('-' : abbr : abbrs) ->
+                  case getAPCData apcData of
+                    ArgsParseControl ArgsInterspersed _ ->
+                      (aas, (am, posn, rest ++ ["--" ++ name]))
+                    _ -> 
+                      parseError usage
+                        ("unknown argument --" ++ name)
+          ('-' : abbr : abbrs)
+            | isJust (Map.lookup abbr abbr_hash) ||
+              apcDash (getAPCData apcData) == ArgsHardDash ->
               case Map.lookup abbr abbr_hash of
                 Just ad ->
-                  let p@(args', state') = peel ['-', abbr] ad aas
+                  let (args', am') = peel ['-', abbr] ad aas
+                      state' = (am', posn, rest)
                   in case abbrs of
-                    [] -> p
+                    [] -> (args', state')
                     ('-' : _) -> parseError usage
                                  ("bad internal '-' in argument " ++ aa)
                     _ -> (['-' : abbrs] ++ args', state')
                 Nothing ->
-                    case acomplete of
+                    case apcComplete $ getAPCData apcData of
                       ArgsInterspersed ->
-                          (['-' : abbrs] ++ aas,
-                           (am, posn, rest ++ [['-', abbr]]))
+                          (aas,
+                           (am, posn, rest ++ ['-' : abbr : abbrs]))
                       _ -> parseError usage
                            ("unknown argument -" ++ [abbr])
-          aa -> case posn of
-                  (ad@(Arg { argData = Just adata }) : ps) ->
-                          let (argl', (am', _, rest')) =
-                                  peel_process (dataArgName adata) ad av
-                          in (argl', (am', ps, rest'))
-                  [] -> case acomplete of
-                          ArgsComplete -> parseError usage
-                                          ("unexpected argument " ++ aa)
-                          _ -> (aas, (am, [], rest ++ [aa]))
+          _ ->
+            case posn of
+              (p : ps) ->
+                let (_, req_posn) = partition arg_optional posn in
+                case length av - length req_posn of
+                  n_extra | n_extra > 0 || (n_extra == 0 && arg_required p) ->
+                    let (args', am') = peel (dataArgName $ fromJust $ 
+                                             argData p) p av in
+                    (args', (am', ps, rest))
+                  0 -> (av, (am, ps, rest))
+                  _ -> parseError usage 
+                         "missing required positional argument(s)"
+              [] -> ([], (am, [], rest ++ av))
         where
           add_entry s m (k, a) =
               case Map.member k m of
                 False -> Map.insert k a m
                 True -> parseError usage ("duplicate argument " ++ s)
-          peel name ad@(Arg { argData = Nothing, argIndex = index }) argl =
+          peel name (Arg { argData = Nothing, argIndex = index }) argl =
               let am' = add_entry name am (index, ArgvalFlag)
-              in (argl, (am', posn, rest))
+              in (argl, am')
           peel name (Arg { argData = Just (DataArg {}) }) [] =
               parseError usage (name ++ " is missing its argument")
-          peel name ad argl = peel_process name ad argl
-          peel_process name
-               ad@(Arg { argData = Just (DataArg {
-                                     dataArgArgtype = atype }),
-                         argIndex = index })
-               (a : argl) =
-                 let read_arg constructor kind =
-                         case reads a of
-                           [(v, "")] -> constructor v
-                           _ -> parseError usage ("argument " ++
-                                                   a ++ " to " ++ name ++
-                                                   " is not " ++ kind)
-                     v = case atype of
-                           ArgtypeString _ -> ArgvalString a
-                           ArgtypeInteger _ -> read_arg ArgvalInteger
-                                                        "an integer"
-                           ArgtypeInt _ -> read_arg ArgvalInt "an int"
-                           ArgtypeDouble _ -> read_arg ArgvalDouble "a double"
-                           ArgtypeFloat _ -> read_arg ArgvalFloat "a float"
-                     am' = add_entry name am (index, v)
-                 in (argl, (am', posn, rest))
+          peel name (Arg { argData = 
+                                 Just (DataArg { dataArgArgtype = atype }),
+                              argIndex = index })
+              (a : argl) =
+                let v = case atype of
+                          ArgtypeString _ -> ArgvalString a
+                          ArgtypeInteger _ -> read_arg ArgvalInteger
+                                                       "an integer"
+                          ArgtypeInt _ -> read_arg ArgvalInt "an int"
+                          ArgtypeDouble _ -> read_arg ArgvalDouble "a double"
+                          ArgtypeFloat _ -> read_arg ArgvalFloat "a float"
+                        where
+                          read_arg constructor kind =
+                            case reads a of
+                              [(val, "")] -> constructor val
+                              _ -> parseError usage ("argument " ++
+                                                     a ++ " to " ++ name ++
+                                                     " is not " ++ kind)
+                    am' = add_entry name am (index, v)
+                in (argl, am')
 
 
 -- |Most of the time, you just want the environment's
 -- arguments and are willing to live in the IO monad.
 -- This version of `parseArgs` digs the pathname and arguments
 -- out of the system directly.
-parseArgsIO :: (Show a, Ord a) =>
-               ArgsComplete  -- ^Degree of completeness of parse.
+parseArgsIO :: (Show a, Ord a, APCData b) =>
+               b             -- ^Degree of completeness of parse.
             -> [ Arg a ]     -- ^Argument descriptions.
             -> IO (Args a)   -- ^Argument parse results.
-parseArgsIO acomplete argd = do
+parseArgsIO apcData argd = do
   argv <- getArgs
   pathname <- getProgName
-  return (parseArgs acomplete argd pathname argv)
+  return (parseArgs apcData argd pathname argv)
 
 
 -- |Check whether a given optional argument was supplied. Works on all types.
@@ -545,19 +584,21 @@ class ArgType b where
            => Args a    -- ^Parsed arguments.
            -> a         -- ^Index of argument to be retrieved.
            -> b   -- ^Argument value.
-    getRequiredArg args index =
-        case getArg args index of
+    getRequiredArg ads index =
+        case getArg ads index of
           Just v -> v
           Nothing -> error ("internal error: required argument "
                           ++ show index ++ "not supplied")
 
+getArgPrimitive :: Ord a => (Argval -> Maybe b) -> Args a -> a -> Maybe b
 getArgPrimitive decons (Args { args = ArgRecord am }) k =
-    case Map.lookup k am of
-      Just v -> Just (decons v)
-      Nothing -> Nothing
+  Map.lookup k am >>= decons
+
+instance ArgType () where
+  getArg = getArgPrimitive (\ArgvalFlag -> return ())
 
 instance ArgType ([] Char) where
-  getArg = getArgPrimitive (\(ArgvalString s) -> s)
+  getArg = getArgPrimitive (\(ArgvalString s) -> return s)
 
 -- |[Deprecated]  Return the `String` value, if any, of the given argument.
 getArgString :: (Show a, Ord a) =>
@@ -567,7 +608,7 @@ getArgString :: (Show a, Ord a) =>
 getArgString = getArg
 
 instance ArgType Integer where
-  getArg = getArgPrimitive (\(ArgvalInteger i) -> i)
+  getArg = getArgPrimitive (\(ArgvalInteger i) -> return i)
 
 -- |[Deprecated] Return the `Integer` value, if any, of the given argument.
 getArgInteger :: (Show a, Ord a) =>
@@ -577,7 +618,7 @@ getArgInteger :: (Show a, Ord a) =>
 getArgInteger = getArg
 
 instance ArgType Int where
-  getArg = getArgPrimitive (\(ArgvalInt i) -> i)
+  getArg = getArgPrimitive (\(ArgvalInt i) -> return i)
 
 -- |[Deprecated] Return the `Int` value, if any, of the given argument.
 getArgInt :: (Show a, Ord a) =>
@@ -587,7 +628,7 @@ getArgInt :: (Show a, Ord a) =>
 getArgInt = getArg
 
 instance ArgType Double where
-  getArg = getArgPrimitive (\(ArgvalDouble i) -> i)
+  getArg = getArgPrimitive (\(ArgvalDouble i) -> return i)
 
 -- |[Deprecated] Return the `Double` value, if any, of the given argument.
 getArgDouble :: (Show a, Ord a) =>
@@ -597,7 +638,7 @@ getArgDouble :: (Show a, Ord a) =>
 getArgDouble = getArg
 
 instance ArgType Float where
-  getArg = getArgPrimitive (\(ArgvalFloat i) -> i)
+  getArg = getArgPrimitive (\(ArgvalFloat i) -> return i)
 
 -- |[Deprecated] Return the `Float` value, if any, of the given argument.
 getArgFloat :: (Show a, Ord a) =>
@@ -612,10 +653,9 @@ newtype ArgFileOpener = ArgFileOpener {
     }
 
 instance ArgType ArgFileOpener where
-    getArg args index =
-        case getArg args index of
-          Nothing -> Nothing
-          Just s -> Just (ArgFileOpener { argFileOpener = openFile s })
+    getArg ads index =
+        getArg ads index >>= 
+          (\s -> return $ ArgFileOpener { argFileOpener = openFile s })
 
 -- |[Deprecated] Treat the `String` value, if any, of the given argument as
 -- a file handle and try to open it as requested.
@@ -625,8 +665,8 @@ getArgFile :: (Show a, Ord a) =>
            -> IOMode              -- ^IO mode the file should be opened in.
            -> IO (Maybe Handle)   -- ^Handle of opened file, if the argument
                                   -- was present.
-getArgFile args k m =
-  case getArg args k of
+getArgFile ads k m =
+  case getArg ads k of
     Just fo -> (do h <- argFileOpener fo m; return (Just h))
     Nothing -> return Nothing
 
@@ -641,8 +681,8 @@ getArgStdio :: (Show a, Ord a) =>
             -> IOMode      -- ^IO mode the file should be opened in.
                            -- Must not be `ReadWriteMode`.
             -> IO Handle   -- ^Appropriate file handle.
-getArgStdio args k m =
-    case getArg args k of
+getArgStdio ads k m =
+    case getArg ads k of
       Just s -> openFile s m
       Nothing ->
           case m of
@@ -669,4 +709,4 @@ baseName s =
 
 -- |Generate a usage error with the given supplementary message string.
 usageError :: (Ord a) => Args a -> String -> b
-usageError args msg = error (argsUsage args ++ "\n" ++ msg)
+usageError ads msg = error (argsUsage ads ++ "\n" ++ msg)
